from rabbitmq_client.common_defs import Printable


"""
Queue Properties
"""


AUTO_GEN_QUEUE_NAME = ""


"""
Consume work item types
"""


class Consumer(Printable):
    """
    Base class for consumers, all consumes have a queue name. This is not set
    until queue name is known in case of fanout subscriptions (Subscription
    work item types).
    """

    def set_queue_name(self, queue_name):
        """
        Setter for queue_name

        :param queue_name: new queue_name
        :return: new consumer_tag
        """
        self.queue_name = queue_name
        return self.queue_name


class Subscription(Consumer):
    """
    A Subscription work item.

    Queue name will be set after queue declaration. Subscriptions uses the
    fanout exchange type, and a queue is generated with a random name to
    receive exchange-bound messages on. This consume work item will receive a
    set queue name after its queue has been declared, so it is possible to find
    out the queue name when receiving the ConsumeOk from the consumer
    connection.
    """

    def __init__(self, topic):
        """
        :param str topic: topic to subscribe to
        """
        self.topic = topic


class RPCServer(Consumer):
    """
    An RPC Server work item.
    """

    def __init__(self, queue_name):
        """
        :param queue_name: name of request queue to declare
        """
        self.queue_name = queue_name


class RPCClient(Consumer):
    """
    An RPC Client work item.
    """

    def __init__(self, queue_name):
        """
        :param queue_name: name of response queue to declare
        """
        self.queue_name = queue_name


class CommandQueue(Consumer):
    """
    A Command Queue work item.
    """

    def __init__(self, queue_name):
        """
        :param queue_name: name for the command queue
        :type queue_name: str
        """
        self.queue_name = queue_name


"""
Monitoring thread interpretable messages, mostly feedback from the Consumer
connection.
"""


class ConsumeOk(Printable):
    """
    Confirmation of a completed consume
    """

    def __init__(self, consumer_tag, consume):
        """
        :param consumer_tag: consumer tag used for cancellation
        :param consume: consume issued
        """
        self.consumer_tag = consumer_tag
        self.consume = consume


class ConsumedMessage(Printable):
    """
    A consumed message.
    """

    def __init__(self,
                 message,
                 exchange,
                 routing_key,
                 correlation_id,
                 reply_to):
        """
        :param message: message received
        :param exchange: exchange received on
        :param routing_key: routing key received on
        :param correlation_id: ID for RPC request/response
        :param reply_to: reply queue for RPC request
        """
        self.message = message
        self.exchange = exchange if exchange else ""
        self.routing_key = routing_key if routing_key else ""
        self.correlation_id = correlation_id if correlation_id else None
        self.reply_to = reply_to if reply_to else None


class StopConsumer(Printable):
    """
    Tell the monitoring thread to stop, posted not from the consumer
    connection, but a Consumer class instance, when stop is called.
    """
    pass



import uuid
import logging

from threading import Event

from .common_defs import Printable
from .consumer_defs import ConsumedMessage


LOGGER = logging.getLogger(__name__)

RPC_REPLY_PREFIX = "RPC-REPLY-"

RPC_DEFAULT_REPLY = b'NONE'


class RPCResponse(Printable):
    """
    RPC response object spec.
    """

    blocker: Event  # used to wait for response
    response: bytes  # defaults to RPC_DEFAULT_REPLY if no response in time

    def __init__(self):
        """"""
        self.blocker = Event()
        self.response = RPC_DEFAULT_REPLY


class RMQRPCHandler:
    """
    RPC handler which uses the consumer and producer instances to handle RPC
    requests and responses.
    """

    def __init__(self, consumer, producer):
        """
        :param consumer: consumer instance
        :param producer: producer instanse
        """
        LOGGER.debug("__init__")

        self._consumer = consumer
        self._producer = producer
        self._response_queue_name = None
        self._request_queue_name = None
        self._rpc_request_callback = None
        self._pending_requests = dict()

    def start(self):
        """
        Not needed, yet.
        """
        pass

    def stop(self):
        """
        Not needed, yet.
        """
        pass

    def enable_rpc_server(self, rpc_queue_name, rpc_request_callback):
        """
        Enables an RPC server, which can accept requests and respond to them.
        The RPC handler will subscribe to the queue name and expect the
        provided callback to RETURN a value which it can reply with.

            rpc_request_callback(message: bytes) -> bytes

         !!! NOTE The importance of the supplied callback to RETURN bytes. !!!

        :param rpc_queue_name: name of RPC request queue to subscribe to
        :param rpc_request_callback: callback to issue requests to
        """
        LOGGER.debug("enable_rpc_server")

        if self._request_queue_name:
            LOGGER.warning("enable_rpc_server an RPC server has"
                           "already been declared")
            return

        self._request_queue_name = rpc_queue_name
        self._rpc_request_callback = rpc_request_callback
        self._consumer.rpc_server(self._request_queue_name,
                                  self.handle_rpc_request)

    def is_rpc_server_ready(self) -> bool:
        """
        Checks if the RPC server is ready, meaning it is consuming on the RPC
        server queue.

        :return: True if ready
        """
        LOGGER.debug("is_rpc_server_ready")

        # If no request queue exists, definitively not ready.
        if not self._request_queue_name:
            return False

        return self._consumer.is_rpc_consumer_ready(self._request_queue_name)

    def enable_rpc_client(self):
        """
        Enables the client to act as an RPC client. This will establish a reply
        queue to receive responses to sent RPC requests.
        """
        LOGGER.debug("enable_rpc_client")

        if self._response_queue_name:
            return

        self._response_queue_name = RPC_REPLY_PREFIX + str(uuid.uuid1())
        self._consumer.rpc_client(self._response_queue_name,
                                  self.handle_rpc_response)

    def is_rpc_client_ready(self) -> bool:
        """
        Check if the RPC client is ready, meaning it is consuming on the RPC
        client's reply queue.

        :return: True if ready
        """
        LOGGER.debug("is_rpc_client_ready")

        # If no response queue exists, definitively not ready.
        if not self._response_queue_name:
            return False

        return self._consumer.is_rpc_consumer_ready(self._response_queue_name)

    def rpc_call(self, receiver, message) -> bytes:
        """
        NOTE! Must enable_rpc_client before making calls to this function.

        Make a synchronous call to an RPC server.

        :param str receiver: name of the RPC server to send the request to
        :param bytes message: message to send to the RPC server

        :return bytes answer: response message from the RPC server
        """
        LOGGER.debug("rpc_call")

        corr_id = str(uuid.uuid1())

        response = RPCResponse()
        self._pending_requests.update({corr_id: response})

        self._producer.rpc_request(receiver,
                                   message,
                                   corr_id,
                                   self._response_queue_name)

        LOGGER.debug("rpc_call blocking waiting for response")

        response.blocker.wait(timeout=2.0)
        if self._pending_requests.get(corr_id):
            LOGGER.info("rpc_call timed out waiting for a response")
            self._pending_requests.pop(corr_id)
        else:
            LOGGER.info("rpc_call got response: {}".format(response))

        return response.response

    def rpc_cast(self, receiver, message, callback):
        """
        NOTE! Must enable_rpc_client before making calls to this function.

        Make an asynchronous call to an RPC server.

        :param receiver: name of the RPC server to send the request to
        :param message: message to send to the RPC server
        :param callback: callback for when response is gotten
        """
        raise NotImplementedError

    def handle_rpc_request(self, message: ConsumedMessage):
        """
        Handler for an incoming RPC request.

        :param message: consumed RPC request
        """
        LOGGER.debug("handle_rpc_request request: {}".format(message))

        answer = self._rpc_request_callback(message.message)

        self._producer.rpc_response(message.reply_to,
                                    answer,
                                    message.correlation_id)

    def handle_rpc_response(self, message: ConsumedMessage):
        """
        Handler for an incoming RPC response.

        :param message: consumed RPC response
        """
        LOGGER.debug(f"handle_rpc_response response: {message}")

        response: RPCResponse = \
            self._pending_requests.pop(message.correlation_id)
        response.response = message.message
        response.blocker.set()



import logging

from logging.handlers import QueueListener

from multiprocessing import Queue as IPCQueue


TOP_LOGGER_NAME = "rabbitmq_client"

_log_manager = None


def initialize_log_manager(log_level=None):
    """
    Initializes logging for the RabbitMQ client. If no log level is provided,
    warnings and up will be printed to console, no null handler is used.

    :param log_level: log level to use
    """
    global _log_manager
    _log_manager = LogManager(log_level)

    if log_level is not None:
        _log_manager.start()


def get_log_manager():
    """
    Interface for getting the log manager instance of the module.

    :return: log manager or None
    """
    return _log_manager if _log_manager is not None else None


def set_process_log_handler(queue, log_level):
    """
    Sets up the processes "root" logger. Well, at least the root of the
    rabbitmq_client project. Queue will be None in case logging is deactivated.

    The idea is that this function is called from EACH started process, so that
    each process has a rabbitmq_client logger configured and a handler which
    puts all log entires on the process shared queue for the log manager to
    consume.

    :param queue: multiprocessing queue to instantiate the QueueHandler with
    :param log_level: log level that the logger and associated handler should
                      handle
    """
    # Found in test case when running them one after the other. If one test
    # case has enabled logging, the rabbitmq_client logger has an assigned
    # QueueHandler referring to an old IPC Queue.

    # This also works as a way of clearing all handlers when a new process is
    # forked.
    logger = logging.getLogger(TOP_LOGGER_NAME)
    logger.handlers = []

    if queue is not None:
        logger.setLevel(log_level)

        handler = logging.handlers.QueueHandler(queue)
        handler.setLevel(log_level)

        logger.addHandler(handler)


class LogManager:
    """
    This class makes sure that all the processes spawned by the RMQClient can
    use a centralized logging solution. This class' member queue is monitored
    for log messages that a process may want to send to file, which the
    LogHandler then takes care of.
    """

    def __init__(self, log_level, filemode='w'):
        """
        Initializes the log handler by creating a process shared queue.

        :param log_level: sets the log level for the log manager
        :type  log_level: logging.DEBUG | logging.INFO | logging.WARNING |
                          logging.ERROR | logging.CRITICAL
        :param filemode: either write or append. Write will overwrite all
                         previous contents.
        :type  filemode: 'w' | 'a'
        """
        self._log_level = log_level

        if self.log_level is not None:

            self.file_handler = logging.FileHandler(
                "rabbitmq_client.log", mode=filemode
            )
            self.file_handler.setLevel(log_level)
            # Padding log level name to 8 characters, CRITICAL is the longest,
            # centered log level by '^'.
            formatter = logging.Formatter(fmt="{asctime} {levelname:^8} "
                                              "{module} {message}",
                                          style="{",
                                          datefmt="%d/%m/%Y %H:%M:%S")
            self.file_handler.setFormatter(formatter)

            self._log_queue = IPCQueue()

            self._listener = QueueListener(
                self._log_queue, self.file_handler, respect_handler_level=True
            )

    def start(self):
        """
        Starts the log manager by starting the listener on the log queue.
        """
        self._listener.start()

    def stop(self):
        """Stops the log manager."""
        if self.log_level is not None:
            self._listener.stop()

            # To avoid resource warnings, the file handler must be explicitly
            # closed.
            self.file_handler.close()

            # flush and close queue
            while not self._log_queue.empty():
                self._log_queue.get()
            self._log_queue.close()
            # In order for client.stop() to be reliable and consistent, ensure
            # thread stop.
            self._log_queue.join_thread()

    @property
    def log_queue(self):
        """
        Getter for _log_queue.

        :return: the manager's log queue
        """
        if self._log_level is not None:
            return self._log_queue
        else:
            return None


    @property
    def log_level(self):
        """
        Getter for log level.

        :return: the log manager's set log level
        """
        return self._log_level



import pika
import logging

from abc import ABCMeta, abstractmethod


LOGGER = logging.getLogger(__name__)


class RMQConnection(metaclass=ABCMeta):
    """
    Abstract class implementing the basics of a RabbitMQ server connection.
    This class does not meddle with channels as the handling of a channel may
    differ between consumers and producers. The RMQConnection class therefore
    only handles the connection (pika.SelectConnection) object itself.

    Subclasses inheriting from RMQConnection have to override the
    on_connection_open(connection: SelectConnection) function to take over once
    the connection has been established. At this point, it is possible to start
    creating channels. Subclasses also need to override
    on_connection_closed(connection: SelectConnection, reason) since consumers
    and producers differ quite a bit in how they handle their connection and
    channels in event of an error.
    """

    _connection_parameters: pika.ConnectionParameters
    _connection: pika.SelectConnection

    _closing: bool

    def __init__(self,
                 connection_parameters):
        """
        Initializes the RMQ connection with connection parameters and the
        general state of the RMQConnection adapter.
        :param connection_parameters: pika.ConnectionParameters or None
        """
        LOGGER.debug("__init__ of connection")

        if connection_parameters:
            self._connection_parameters = connection_parameters
        else:
            # Default credentials: guest/guest
            # Default host: localhost
            # Default vhost: /
            # Default port: 5672
            self._connection_parameters = pika.ConnectionParameters()

        self._connection = pika.SelectConnection(
            parameters=self._connection_parameters,
            on_open_callback=self.on_connection_open,
            on_close_callback=self.on_connection_closed
        )

        self._closing = False

    def connect(self):
        """
        Initiates the connection to the RabbitMQ server by starting the
        connection's IO-loop. Starting the IO-loop will connect to the RabbitMQ
        server as configured in __init__.
        """
        LOGGER.debug("connect in connection")

        self._connection.ioloop.start()

    @abstractmethod
    def on_connection_open(self, _connection):
        """
        Callback upon opened connection. Subclasses shall override this
        function in order to be notified when a connection has been established
        in order for them to know when they are able to create channels.

        :param pika.SelectConnection _connection: connection that was opened
        """
        pass

    @abstractmethod
    def on_connection_closed(self, _connection, reason):
        """
        Callback upon closed connection. Subclasses shall override this
        function in order to be notified when a connection has been closed.

        :param pika.SelectConnection _connection: connection that was closed
        :param Exception reason: reason for closing
        """
        pass

    def disconnect(self):
        """
        Disconnects from the RabbitMQ server by calling close() on the
        connection object. This operation should result in on_connection_closed
        being invoked once the connection has been closed, allowing for further
        handling of either gracefully shutting down or re-connecting.
        """
        LOGGER.debug("disconnect in connection")

        if not self._closing:
            self._closing = True
            self._connection.close()
            return

    def finalize_disconnect(self):
        """
        Shall be called once the connection has been closed to stop the
        IO-Loop.
        """
        LOGGER.debug("finalizing disconnect")

        self._connection.ioloop.stop()



import logging

from rabbitmq_client import log

from rabbitmq_client.log import LogManager
from rabbitmq_client.rpc import RMQRPCHandler
from rabbitmq_client.consumer import RMQConsumer
from rabbitmq_client.producer import RMQProducer


LOGGER = logging.getLogger(__name__)


class RMQClient:
    """
    Provides an easy to use interface for applications wanting to leverage the
    power of RabbitMQ and the awesome pika Python library.
    """

    def __init__(self,
                 log_level=None,
                 connection_parameters=None):
        """
        :param log_level: sets the log level of the RMQClient, None being no
                          logging at all.
        :param connection_parameters: connection parameters to the RMQ server
        :type connection_parameters: pika.ConnectionParameters
        """
        log.initialize_log_manager(log_level=log_level)

        log_manager: LogManager = log.get_log_manager()
        # Client process log handler is set here!
        log.set_process_log_handler(log_manager.log_queue, log_level)

        self._consumer = RMQConsumer(connection_parameters)
        self._producer = RMQProducer(connection_parameters)

        self._rpc_handler = RMQRPCHandler(self._consumer,
                                          self._producer)

    def start(self):
        """
        Starts the RMQClient by starting its subsequent consumer and producer
        connections.
        """
        LOGGER.info("start")

        self._consumer.start()
        self._producer.start()

    def stop(self):
        """
        Stops the RMQ client and its child applications/processes.
        """
        LOGGER.info("stop")

        self._consumer.stop()
        self._producer.stop()

        # Logging messages from consumer and producer are lost if logging is
        # not closed last.
        log_manager = log.get_log_manager()
        if log_manager:
            log_manager.stop()

    def subscribe(self, topic, callback):
        """
        Subscribes to the input topic. A message received for the given topic
        will result in the given callback being called.

            callback(message: bytes)

        The function returns immediately but will dispatch the subscription job
        to another process, meaning that the actual subscribing will take place
        shortly after the function returns.

        :param str topic: topic to subscribe to
        :param callable callback: callback on message received
        """
        LOGGER.info(f"subscribe topic: {topic} callback: {callback}")

        # dynamically add a subscription to a topic
        self._consumer.subscribe(topic, callback)

    def is_subscribed(self, topic) -> bool:
        """
        Checks if a subscription has been activated, meaning RMQ has confirmed
        with a ConsumeOk that a consume is active.

        :param topic: topic to check
        :return: true if active
        """
        LOGGER.info(f"is_subscribed topic: {topic}")

        return self._consumer.is_subscribed(topic)

    def publish(self, topic, message):
        """
        Publishes a message on the given topic. The publising work is
        dispatched to another process, meaning publishing probably has not
        completed by the time this function returns.

        :param str topic: topic to publish on
        :param bytes message: message to publish
        """
        LOGGER.info(f"publish to topic: {topic} message: {message}")
        # publishes a message to the provided topic
        self._producer.publish(topic, message)

    def enable_rpc_server(self, rpc_queue_name, rpc_request_callback):
        """
        Enables an RPC server for the supplied RPC queue name. The client will
        subscribe to messages on the supplied queue and relay incoming requests
        to the supplied callback.

            rpc_request_callback(message: bytes) -> bytes

         !!! NOTE The importance of the supplied callback to RETURN bytes. !!!

        :param rpc_queue_name: string name of the RPC request queue
        :param rpc_request_callback: callback called upon incoming request
        """
        LOGGER.info("enable_rpc_server rpc_queue_name: {} "
                    "rpc_request_callback: {}"
                    .format(rpc_queue_name, rpc_request_callback))
        self._rpc_handler.enable_rpc_server(rpc_queue_name,
                                            rpc_request_callback)

    def is_rpc_server_ready(self) -> bool:
        """
        Checks if the RPC server is ready, meaning it is consuming on the RPC
        server queue.

        :return: True if ready
        """
        LOGGER.info("is_rpc_server_ready")

        return self._rpc_handler.is_rpc_server_ready()

    def enable_rpc_client(self):
        """
        Enables the client to act as an RPC client. This will make sure that
        the client can handle sending RPC requests.
        """
        LOGGER.info("enable_rpc_client")
        self._rpc_handler.enable_rpc_client()

    def is_rpc_client_ready(self) -> bool:
        """
        Check if the RPC client is ready, meaning it is consuming on the RPC
        client's reply queue.

        :return: True if ready
        """
        LOGGER.info("is_rpc_client_ready")

        return self._rpc_handler.is_rpc_client_ready()

    def rpc_call(self, receiver, message) -> bytes:
        """
        NOTE! Must enable_rpc_client before making calls to this function.

        Make a synchronous call to an RPC server.

        :param str receiver: name of the RPC server to send the request to
        :param bytes message: message to send to the RPC server

        :return bytes answer: response message from the RPC server
        """
        LOGGER.info("rpc_call receiver: {} message: {}"
                    .format(receiver, message))
        return self._rpc_handler.rpc_call(receiver, message)

    def rpc_cast(self, receiver, message, callback):
        """
        NOTE! Must enable_rpc_client before making calls to this function.

        Make an asynchronous call to an RPC server.

        :param receiver: name of the RPC server to send the request to
        :param message: message to send to the RPC server
        :param callback: callback for when response is gotten
        """
        raise NotImplementedError

    def command_queue(self, queue_name, callback):
        """
        Declare and consume from the named queue name. The declared queue will
        have the following properties:

        1. Durable
        2. Exclusive consumer (only 1 consumer allowed)

        Command queues are intended as a single point of communication towards
        a given service, identified by its queue name. Commands sent shall be
        handled in order, therefore there is queue durability and no more than
        one consumer is allowed at any point in time.

        NOTE! Ensure the consuming service is ready to receive requests before
        declaring a command queue, as there already may be messages waiting
        after a restart. If not, expect cyclic crashes.

        :param queue_name: name of the command queue to be declared and
                           consumed from
        :type queue_name: str
        :param callback: function to call on message reception
        :type callback: callable
        """
        LOGGER.info(f"Command queue requested: {queue_name}")

        self._consumer.command_queue(queue_name, callback)

    def command(self, command_queue, command):
        """
        Send a command to a command queue.

        :param command_queue: name of the command queue to send the command to
        :type command_queue: str
        :param command: command to send to command queue
        :type command: bytes
        """
        LOGGER.info(f"command {command} to command queue: {command_queue}")

        self._producer.command(command_queue, command)



import logging

from multiprocessing import Queue as IPCQueue, Process

from rabbitmq_client import log

from .log import LogManager
from .producer_connection import create_producer_connection
from .producer_defs import Publish, RPCRequest, RPCResponse, Command


LOGGER = logging.getLogger(__name__)


class RMQProducer:
    """
    Interface for producer-related operations towards a RabbitMQ server.

    =================
    Pub/Sub
    =================
    Supported, not much to say really. Will publish to named exchange and use
    'fanout' to distribute messages to all listening queues.

    =================
    RPC
    =================
    Acts more as an interface towards the producer connection for the
    RPCHandler than actually doing something useful. Only instantiates work
    items and puts them on the work queue.
    """

    def __init__(self,
                 connection_parameters):
        """
        :param connection_parameters: pika.ConnectionParameters or None
        """
        LOGGER.debug("__init__")

        self._work_queue = IPCQueue()

        log_manager: LogManager = log.get_log_manager()

        self._connection_process = Process(
            target=create_producer_connection,
            args=(
                connection_parameters,
                self._work_queue,

                # This is fine since we're still in the same process!
                log_manager.log_queue,
                log_manager.log_level
            )
        )

    def start(self):
        """
        Starts the producer's connection process in order to be able to publish
        messages. The connection process is maintained in another process, the
        work queue passed along to the new process is process shared to allow
        for the controlling process to issue commands to the connection
        process, for example to publish messages.
        """
        LOGGER.info("start")

        self._connection_process.start()

    def stop(self):
        """
        Stops the RMQConsumer, tearing down the RMQProducerConnection process.
        """
        LOGGER.info("stop")

        self.flush_and_close_queues()

        self._connection_process.terminate()
        self._connection_process.join(timeout=2)

    def flush_and_close_queues(self):
        """
        Flushed process shared queues in an attempt to stop background threads.
        """
        while not self._work_queue.empty():
            self._work_queue.get()
        self._work_queue.close()
        # In order for client.stop() to be reliable and consistent, ensure
        # thread stop.
        self._work_queue.join_thread()

    def publish(self, topic, message):
        """
        Publishes a message on the supplied topic.

        :param str topic: topic to publish on
        :param bytes message: message content
        """
        LOGGER.debug("publish")

        self._work_queue.put(Publish(topic, message))

    def rpc_request(self, receiver, message, correlation_id, reply_to):
        """
        Interface for sending an RPC request to the producer connection.

        :param receiver: receiving RPC queue name
        :param message: contents to send
        :param correlation_id: identifies the request
        :param reply_to: response queue name
        """
        LOGGER.debug(f"rpc_request receiver: {receiver} message: {message} "
                     f"correlation_id: {correlation_id} reply_to: {reply_to}")

        self._work_queue.put(RPCRequest(receiver,
                                        message,
                                        correlation_id,
                                        reply_to))

    def rpc_response(self, receiver, message, correlation_id):
        """
        Interface for sending an RPC response to the producer connection.

        :param receiver: receiving RPC response queue name
        :param message: contents to send
        :param correlation_id: identifies which request the response belongs to
        """
        LOGGER.debug(f"rpc_reply receiver: {receiver} message: {message} "
                     f"correlation_id: {correlation_id}")

        self._work_queue.put(RPCResponse(receiver,
                                         message,
                                         correlation_id))

    def command(self, command_queue, command):
        """
        Send command to specified command queue.

        :param command_queue: name of the command queue to send the command to
        :type command_queue: str
        :param command: command to send to command queue
        :type command: bytes
        """
        LOGGER.debug(f"command {command} to {command_queue}")

        self._work_queue.put(Command(command_queue, command))



import functools
import logging

from .consumer_defs import Subscription, RPCServer, RPCClient, \
                           ConsumedMessage, ConsumeOk, CommandQueue, \
                           AUTO_GEN_QUEUE_NAME
from .common_defs import EXCHANGE_TYPE_FANOUT


LOGGER = logging.getLogger(__name__)


class RMQConsumerChannel:
    """
    Defines handling of a consumer channel. Declarations of queues and
    exchanges, consuming and error handling.
    """

    def __init__(self, consumed_messages):
        """
        :param consumed_messages: IPC queue used to relay consumed messages
                                  between processes
        """
        LOGGER.debug("__init__")

        self._consumed_messages = consumed_messages

        self._channel = None
        self._open = False

    def open_channel(self, connection, notify_callback):
        """
        Opens a channel for the parameter connection.

        :param connection: current connection object
        :param notify_callback: callback to notify once channel is ready for
                                work
        """
        LOGGER.debug("open_channel")

        cb = functools.partial(self.on_channel_open,
                               notify_callback=notify_callback)

        if not self._open:
            connection.channel(on_open_callback=cb)

    def on_channel_open(self, channel, notify_callback=None):
        """
        Callback for channel open.

        :param channel: opened channel
        :param notify_callback: callback to notify once channel is ready for
                                work
        """
        LOGGER.info("on_channel_open channel: {}".format(channel))

        self._open = True

        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)

        notify_callback()

    def on_channel_closed(self, _channel, reason):
        """
        Callback for channel closed.

        :param _channel: closed channel
        :param reason: reason channel was closed (exception)
        """
        LOGGER.info(f"on_channel_closed channel: {_channel} reason: {reason}")

        self._open = False

    def handle_consume(self, consume):
        """
        Initiates a new consumer for the channel.

        :param consume: consumer information needed to establish the new
                        consumer
        """
        LOGGER.debug(f"handle_consume consume: {consume}")

        if isinstance(consume, Subscription):
            cb = functools.partial(self.on_exchange_declared,
                                   consume=consume)
            self._channel.exchange_declare(exchange=consume.topic,
                                           exchange_type=EXCHANGE_TYPE_FANOUT,
                                           callback=cb)

        elif isinstance(consume, RPCServer):
            cb = functools.partial(self.on_queue_declared,
                                   consume=consume)
            self._channel.queue_declare(queue=consume.queue_name,
                                        durable=True,  # Survive broker reboot
                                        callback=cb)

        elif isinstance(consume, RPCClient):
            cb = functools.partial(self.on_queue_declared,
                                   consume=consume)
            self._channel.queue_declare(queue=consume.queue_name,
                                        exclusive=True,
                                        callback=cb)

        elif isinstance(consume, CommandQueue):
            cb = functools.partial(self.on_queue_declared,
                                   consume=consume)
            self._channel.queue_declare(queue=consume.queue_name,
                                        durable=True,  # Survive broker reboot
                                        callback=cb)

    def on_exchange_declared(self, _frame, consume):
        """
        Callback for when an exchange has been declared.

        :param pika.frame.Method _frame: message frame
        :param consume: consumer information needed to establish the new
                        consumer
        """
        LOGGER.debug("on_exchange_declared frame: {} consume: {}"
                     .format(_frame, consume))

        cb = functools.partial(self.on_queue_declared,
                               consume=consume)
        self._channel.queue_declare(queue=AUTO_GEN_QUEUE_NAME,
                                    exclusive=True,
                                    callback=cb)

    def on_queue_declared(self, frame, consume):
        """
        Callback for when a queue has been declared.

        :param pika.frame.Method frame: message frame
        :param consume: consumer information needed to establish the new
                        consumer
        """
        LOGGER.debug(f"on_queue_declared frame: {frame} consume: {consume}")

        if isinstance(consume, Subscription):
            consume.set_queue_name(frame.method.queue)
            cb = functools.partial(self.on_queue_bound,
                                   consume=consume)
            self._channel.queue_bind(
                consume.queue_name, consume.topic, callback=cb
            )

        elif isinstance(consume, RPCServer) or \
                isinstance(consume, RPCClient) or \
                isinstance(consume, CommandQueue):
            # No exchange = no need to bind the queue, can go ahead and consume
            # immediately.
            self.consume(consume)

    def on_queue_bound(self, _frame, consume):
        """
        Callback for when a queue has been bound to an exchange.

        :param pika.frame.Method _frame: message frame
        :param consume: consumer information needed to establish the new
                        consumer
        """
        LOGGER.debug("on_queue_bound frame: {} consume: {}"
                     .format(_frame, consume))

        self.consume(consume)

    def consume(self, consume):
        """
        Starts consuming on the parameter queue.

        :param consume: consume action
        """
        LOGGER.info("consume queue_name: {}"
                    .format(consume.queue_name))

        cb = functools.partial(self.on_consume_ok,
                               consume=consume)
        # All consumes so far are exclusive, meaning there can be only one
        # consumer for the given queue.
        self._channel.basic_consume(consume.queue_name,
                                    self.on_message,
                                    exclusive=True,
                                    callback=cb)

    def on_message(self, _channel, basic_deliver, properties, body):
        """
        Callback for when a message is received on a consumed queue.

        :param _channel: channel that the message was received on
        :param pika.spec.Basic.Deliver basic_deliver: method
        :param pika.spec.BasicProperties properties: properties of the message
        :param bytes body: message body
        """
        LOGGER.info(f"on_message method: {basic_deliver} properties: "
                    f"{properties} body: {body}")

        self._consumed_messages.put(
            ConsumedMessage(body,
                            basic_deliver.exchange,
                            basic_deliver.routing_key,
                            properties.correlation_id if properties else None,
                            properties.reply_to if properties else None)
        )

        self._channel.basic_ack(basic_deliver.delivery_tag)

    def on_consume_ok(self, frame, consume):
        """
        Callback for when confirm mode has been activated.

        :param pika.spec.Method frame: message frame
        :param consume: issued consume
        """
        LOGGER.info("on_consume_ok frame: {}".format(frame))

        self._consumed_messages.put(
            ConsumeOk(frame.method.consumer_tag, consume)
        )



import logging

from multiprocessing import Queue as IPCQueue, Process
from threading import Thread

from rabbitmq_client import log
from rabbitmq_client.log import LogManager

from rabbitmq_client.errors import ConsumerAlreadyExists

from rabbitmq_client.consumer_connection import create_consumer_connection
from rabbitmq_client.common_defs import Printable
from rabbitmq_client.consumer_defs import \
    Subscription, ConsumedMessage, ConsumeOk, StopConsumer, RPCServer, \
    RPCClient, CommandQueue


LOGGER = logging.getLogger(__name__)


class Consumer(Printable):
    """
    Used to package what makes up a consumer to the RMQConsumer class. This
    class is instantiated when a new subscription is made.

    The preserve flag is used to subscribe internally in the RMQ client, where
    the raw message format (ConsumedMessage) is used instead of extracting only
    the delivered message.
    """

    def __init__(self,
                 callback,
                 preserve=False,
                 exchange="",
                 routing_key="",
                 consumer_tag=""):
        """
        :param callback: callback to call on new message
        :param preserve: set if you want to preserve original message format:
                         ConsumedMessage instead of only getting bytes.
        :param exchange: subscribed exchange
        :param routing_key: subscribed routing key (queue name in case of
                            default exchange)
        """
        LOGGER.debug("__init__ Consumer")

        self.callback = callback
        self.preserve = preserve
        self.exchange = exchange
        self.routing_key = routing_key
        self.consumer_tag = consumer_tag

    def set_consumer_tag(self, new_consumer_tag):
        """
        Setter for consumer_tag

        :param new_consumer_tag: new consumer_tag
        :return: new consumer_tag
        """
        LOGGER.debug(f"setting consumer tag to: {new_consumer_tag}")

        self.consumer_tag = new_consumer_tag
        return self.consumer_tag


class RMQConsumer:
    """
    Implements an asynchronous consumer for RabbitMQ.

    =================
    Pub/Sub
    =================
    Subscriptions can be added dynamically at any point, even before a
    connection has been established to the RMQ-server. The operation will
    simply be put on hold until exchanges/queues can be declared.

    =================
    RPC
    =================
    To support RPC, subscriptions are added with the preserve flag, and the
    whole ConsumedMessage is forwarded to the RPCHandler. Other than that, the
    RPC support works by subscribing just like any other subscription, only
    that the recepient is the RPCHandler.
    """

    def __init__(self,
                 connection_parameters):
        """
        :param connection_parameters: pika.ConnectionParameters or None
        """
        LOGGER.debug("__init__")

        self._consumers = list()

        self._work_queue = IPCQueue()
        self._consumed_messages = IPCQueue()

        # This is fine since we're still in the same process!
        log_manager: LogManager = log.get_log_manager()

        self._connection_process = Process(
            target=create_consumer_connection,
            args=(
                connection_parameters,
                self._work_queue,
                self._consumed_messages,

                log_manager.log_queue,
                log_manager.log_level
            )
        )

        self._monitoring_thread = Thread(target=self.consume, daemon=True)

    def start(self):
        """
        Starts the RMQConsumer, meaning it is prepared for consuming messages.

        By starting the RMQConsumer, a process is created which will hold an
        RMQConsumerConnection. This function also starts a thread in the
        current process that monitors the consumed_messages queue for incoming
        messages.
        """
        LOGGER.info("start")

        self._connection_process.start()
        self._monitoring_thread.start()

    def stop(self):
        """
        Stops the RMQConsumer, tearing down the RMQConsumerConnection process.
        """
        LOGGER.info("stop")

        self._consumed_messages.put(StopConsumer())
        self._monitoring_thread.join()

        self._connection_process.terminate()
        self._connection_process.join(timeout=2)

    def add_new_consumer(self, new_consumer: Consumer):
        """
        Adds a new Consumer instance to the RMQConsumer's list of consumers, as
        long as the new consumer is not clashing with an old one.

        :param new_consumer: new consumer instance to add
        :raises ConsumerAlreadyExists: raised if the supplied consumer clashes
                                       with an existing one
        """
        matching_consumers = \
            [consumer for consumer in self._consumers
             if consumer.exchange == new_consumer.exchange and
             consumer.routing_key == new_consumer.routing_key]

        if matching_consumers:
            raise ConsumerAlreadyExists("A matching consumer already exists")

        self._consumers.append(new_consumer)

    def consume(self):
        """
        Monitors the consumed_messages queue for any incoming messages.
        """
        LOGGER.debug("consume")
        while True:
            LOGGER.debug("waiting for consumed messages")

            message = self._consumed_messages.get()
            LOGGER.debug("got new consumed message")

            # From Consumer connection
            if isinstance(message, ConsumedMessage):
                self.handle_message(message)
            elif isinstance(message, ConsumeOk):
                self.handle_consume_ok(message)

            # For stopping the monitoring thread
            elif isinstance(message, StopConsumer):
                self.flush_and_close_queues()
                break

    def flush_and_close_queues(self):
        """
        Flushed process shared queues in an attempt to stop background threads.
        """
        LOGGER.debug("flush_and_close_queues")
        while not self._consumed_messages.empty():
            self._consumed_messages.get()
        self._consumed_messages.close()
        # In order for client.stop() to be reliable and consistent, ensure
        # thread stop.
        self._consumed_messages.join_thread()

        while not self._work_queue.empty():
            self._work_queue.get()
        self._work_queue.close()
        # In order for client.stop() to be reliable and consistent, ensure
        # thread stop.
        self._work_queue.join_thread()

    def handle_message(self, message: ConsumedMessage):
        """
        Defines handling for a received message, dispatches the message
        contents to registered callbacks depending on the topic.

        :param ConsumedMessage message: received message
        """
        LOGGER.debug("handle_message message: {}".format(message))

        for consumer in self._consumers:
            if message.exchange == consumer.exchange and \
               message.routing_key == consumer.routing_key:
                if consumer.preserve:  # Internal consumer, forward object
                    consumer.callback(message)
                else:  # Only message content
                    consumer.callback(message.message)

    def handle_consume_ok(self, message: ConsumeOk):
        """
        Defines handling for a received ConsumeOk message, meaning a consume
        has been successfully started for a consumer.

        :param message: contains information about the started consumer
        """
        LOGGER.info(f"handle_consume_ok for: {message.consume.queue_name}")

        if isinstance(message.consume, Subscription):
            # filter out subscription consumer and set consumer_tag
            [consumer.set_consumer_tag(message.consumer_tag)
             for consumer in self._consumers
             if consumer.exchange == message.consume.topic]
        else:  # RPCServer or RPCClient
            [consumer.set_consumer_tag(message.consumer_tag)
             for consumer in self._consumers
             if consumer.routing_key == message.consume.queue_name]

        consumer_state = ""
        for consumer in self._consumers:
            consumer_state = consumer_state + "\n\t{}".format(consumer)

        LOGGER.debug("handle_consume_ok new consumer state: {}"
                     .format(consumer_state))

    def subscribe(self, topic, callback):
        """
        Subscribes to messages sent to the named topic. Messages received on
        this topic will be dispatched to the provided callback.

            callback(message: bytes)

        Updates the internal dictionary topic_callback so that the input
        callback will be called when a message is received for the given topic.

        :param str topic: topic to subscribe to
        :param callable callback: callback on message received
        """
        # 1. Add callback to be called when event on that topic + routing_key
        # 2. Request a subscription on the new topic towards the consumer
        #    connection
        LOGGER.debug("subscribe")

        self.add_new_consumer(Consumer(callback, exchange=topic))
        self._work_queue.put(Subscription(topic))

    def is_subscribed(self, topic):
        """
        Checks if a subscription has been activated, meaning RMQ has confirmed
        with a ConsumeOk that a consume is active.

        :param topic: topic to check
        :return: true if active
        """
        result = False

        for consumer in self._consumers:
            if not consumer.preserve:  # RPC
                if consumer.exchange == topic and \
                   consumer.consumer_tag:  # If not "", ConsumeOK gotten
                    result = True
                    break

        LOGGER.debug("is_subscribed {}".format(result))

        return result

    def rpc_server(self, queue_name, callback):
        """
        Starts an RPC server by issuing a subscribe request to the consumer
        connection.

        The preserve flag ensures that the raw ConsumedMessage object is
        forwarded to the supplied callback function.

        :param queue_name: RPC server request queue name
        :param callback: callback on message received
        """
        LOGGER.debug("rpc_server")

        self.add_new_consumer(Consumer(callback,
                                       preserve=True,
                                       routing_key=queue_name))
        self._work_queue.put(RPCServer(queue_name))

    def rpc_client(self, queue_name, callback):
        """
        Starts an RPC client by issuing a subscribe request to the consumer
        connection.

        The preserve flag ensures that the raw ConsumedMessage object is
        forwarded to the supplied callback function.

        :param queue_name: RPC client response queue name
        :param callback: callback on message received
        """
        LOGGER.debug("rpc_client")

        self.add_new_consumer(Consumer(callback,
                                       preserve=True,
                                       routing_key=queue_name))
        self._work_queue.put(RPCClient(queue_name))

    def is_rpc_consumer_ready(self, server_name) -> bool:
        """
        Checks if the RPC server is ready.

        :return: True if ready
        """
        result = False

        for consumer in self._consumers:
            if consumer.preserve:
                if consumer.routing_key == server_name and \
                   consumer.consumer_tag:
                    result = True
                    break

        LOGGER.debug(f"is_rpc_consumer_ready {result}")

        return result

    def command_queue(self, queue_name, callback):
        """
        Adds a command queue instance to the list of consumers and issues a
        request to the consumer connection to set up the command queue.
        """
        LOGGER.debug(f"command_queue {queue_name}")

        self.add_new_consumer(Consumer(callback, routing_key=queue_name))
        self._work_queue.put(CommandQueue(queue_name))



import signal
import logging

from threading import Thread

from .producer_channel import RMQProducerChannel
from .connection import RMQConnection
from rabbitmq_client import log


LOGGER = logging.getLogger(__name__)


def create_producer_connection(connection_parameters,
                               work_queue,
                               log_queue,
                               log_level):
    """
    Interface function to instantiate and connect a producer connection. This
    function is intended as a target for a new process to avoid having to
    instantiate the RMQProducerConnection outside of the new process' memory
    context.

    :param connection_parameters: pika.ConnectionParameters or None
    :param work_queue: process shared queue used to issue work for the
                       producer connection
    :param log_queue: queue to post log writes to
    :param log_level: log level
    """
    # Configure logging
    log.set_process_log_handler(log_queue, log_level)

    producer_connection = RMQProducerConnection(connection_parameters,
                                                work_queue)
    producer_connection.connect()


class RMQProducerConnection(RMQConnection):
    """
    This class handles a connection to a RabbitMQ server intended for a
    producer entity.

    Messages to be published are posted to a process shared queue which
    is read continuously by a connection process-local thread assigned to
    monitoring the queue.
    """

    def __init__(self,
                 connection_parameters,
                 work_queue):
        """
        Initializes the RMQProducerConnection's work queue and binds signal
        handlers. The work queue can be used to issue commands.

        :param connection_parameters: pika.ConnectionParameters or None
        :param IPCQueue work_queue: process shared queue used to issue work for
                                    the consumer connection
        """
        LOGGER.debug("__init__")

        self._work_queue = work_queue

        self._channel = RMQProducerChannel()

        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.terminate)

        super().__init__(connection_parameters)

    def on_connection_open(self, connection):
        """
        Callback when a connection has been established to the RMQ server.

        :param pika.SelectConnection connection: established connection
        """
        LOGGER.info("on_connection_open connection: {}"
                    .format(connection))

        self._channel.open_channel(connection, self.on_channel_open)

    def on_connection_closed(self, _connection, reason):
        """
        Callback for when a connection is closed.

        :param _connection: connection that was closed
        :param reason: reason for closing
        """
        log_string = "on_connection_closed connection: {} reason: {}" \
            .format(_connection, reason)

        if self._closing:
            LOGGER.info(log_string)
            self.finalize_disconnect()
        else:
            # TODO: reconnect handling goes here
            LOGGER.critical(log_string)
            self.finalize_disconnect()

    def on_channel_open(self):
        """
        Callback for notifying a channel has been completely established.
        """
        LOGGER.debug("on_channel_open")

        self.producer_connection_started()

    def producer_connection_started(self):
        """
        Shall be called when the producer connection has reached a state where
        it is ready to receive and execute work, for instance to publish
        messages.
        """
        LOGGER.info("producer_connection_started")

        thread = Thread(target=self.monitor_work_queue, daemon=True)
        thread.start()

    def monitor_work_queue(self):
        """
        NOTE!

        This function should live in its own thread so that the
        RMQProducerConnection is able to respond to incoming work as quickly as
        possible.

        Monitors the producer connection's work queue and executes from it as
        soon as work is available.
        """
        while True:
            LOGGER.debug("monitor_work_queue waiting for work")

            # Blocking, set block=false to not block
            work = self._work_queue.get()
            LOGGER.debug("got work to do")
            self._channel.handle_produce(work)

    def interrupt(self, _signum, _frame):
        """
        Signal handler for signal.SIGINT.

        :param int _signum: signal.SIGINT
        :param ??? _frame: current stack frame
        """
        LOGGER.debug("interrupt")

        self.disconnect()

    def terminate(self, _signum, _frame):
        """
        Signal handler for signal.SIGTERM.

        :param int _signum: signal.SIGTERM
        :param ??? _frame: current stack frame
        """
        LOGGER.debug("terminate")

        self.disconnect()



"""
Exchange properties
"""


EXCHANGE_TYPE_FANOUT = "fanout"


"""
Generic class info
"""


class Printable:
    """
    Makes the print format like I want it :-).
    """

    def __str__(self):
        return "{} {}".format(self.__class__, self.__dict__)



import functools
import logging

from threading import Lock

from pika.spec import Basic, BasicProperties

from rabbitmq_client.common_defs import EXCHANGE_TYPE_FANOUT
from rabbitmq_client.producer_defs import Publish, RPCRequest, RPCResponse, \
                                          Command, DEFAULT_EXCHANGE


LOGGER = logging.getLogger(__name__)


def max_attempts_reached(work):
    """
    Checks if max attemts have been reached.

    :param work: true if work.attempts > work.MAX_ATTEMPTS
    """
    if work.attempts > work.MAX_ATTEMPTS:
        return True
    return False


class RMQProducerChannel:
    """
    Defines handling of a producer channel, enables confirm mode and keeps
    track of delivery tags with each sent message.
    """

    def __init__(self):
        """"""
        LOGGER.debug("__init__")

        self._expected_delivery_tag = 0
        self._delivery_tag_lock = Lock()
        self._pending_confirm = dict()

        self._channel = None
        self._open = False

    def open_channel(self, connection, notify_callback):
        """
        Opens a channel for the parameter connection.

        :param connection: connection to open channel for
        :param notify_callback: callback to notify once channel has been
                                created
        """
        LOGGER.debug("open_channel")

        cb = functools.partial(self.on_channel_open,
                               notify_callback=notify_callback)

        if not self._open:
            connection.channel(on_open_callback=cb)

    def on_channel_open(self, channel, notify_callback=None):
        """
        Callback for when a channel has been established on the connection.

        :param pika.channel.Channel channel: the opened channel
        :param notify_callback: callback to notify once channel has been
                                created
        """
        LOGGER.info("on_channel_open channel: {}".format(channel))

        self._open = True

        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)

        cb = functools.partial(self.on_confirm_mode_activated,
                               notify_callback=notify_callback)
        self._channel.confirm_delivery(
            ack_nack_callback=self.on_delivery_confirmed,
            callback=cb
        )

    def on_channel_closed(self, channel, reason):
        """
        Callback for when a channel has been closed.

        :param pika.channel.Channel channel: the channel that was closed
        :param Exception reason: exception explaining why the channel was
                                 closed
        """
        LOGGER.info(f"on_channel_closed channel: {channel} reason: {reason}")

        self._open = False

    def on_confirm_mode_activated(self, _frame, notify_callback=None):
        """
        Callback for when confirm mode has been activated.

        :param pika.frame.Method _frame: message frame
        :param notify_callback: callback to notify once channel has been
                                created
        """
        LOGGER.debug(f"on_confirm_mode_activated frame: {_frame}")

        notify_callback()

    def handle_produce(self, produce):
        """
        Handler function for produce work.

        :param produce: work to be done by the producer channel
        :type produce: Publish | RPCRequest | RPCResponse | Command
        """
        LOGGER.debug("handle_produce work: {}".format(produce))

        if max_attempts_reached(produce):
            LOGGER.critical(f"handle_rpc_response max attempts "
                            f"reached for: {produce}, aborting produce")
            return

        if isinstance(produce, Publish):
            cb = functools.partial(self.on_exchange_declared,
                                   produce=produce)
            self._channel.exchange_declare(exchange=produce.topic,
                                           exchange_type=EXCHANGE_TYPE_FANOUT,
                                           callback=cb)

        elif isinstance(produce, RPCRequest):
            cb = functools.partial(self.on_queue_declared,
                                   produce=produce)
            self._channel.queue_declare(queue=produce.receiver,
                                        durable=True,  # Survive broker reboot
                                        callback=cb)

        elif isinstance(produce, RPCResponse):
            # Since just got RPC request, send RPC response without checking
            # for queue declaration. Besides, the RPC reply queue is exclusive.
            self.rpc_response(produce)

        elif isinstance(produce, Command):
            cb = functools.partial(self.on_queue_declared,
                                   produce=produce)
            self._channel.queue_declare(queue=produce.command_queue,
                                        durable=True,  # Survive broker reboot
                                        callback=cb)

    def on_exchange_declared(self, _frame, produce):
        """
        Callback for when an exchange has been declared.

        :param pika.frame.Method _frame: message frame
        :param produce: additional parameter from functools.partial,
                        used to carry the publish object
        :type produce: Publish
        """
        LOGGER.debug(f"on_exchange_declared frame: {_frame}")

        self.publish(produce)

    def on_queue_declared(self, frame, produce):
        """
        Callback for when a queue has been declared.

        :param pika.frame.Method frame: message frame
        :param produce: information about the ongoing produce
        :type produce: RPCRequest | RPCResponse | Command
        """
        LOGGER.debug(f"on_queue_declared frame: {frame} produce: {produce}")

        # RPC responses do not check for queue declaration, nor do publishes.
        if isinstance(produce, RPCRequest):
            self.rpc_request(produce)
        elif isinstance(produce, Command):
            self.command(produce)

    def publish(self, publish):
        """
        Perform a publish operation.

        :param publish: the publish operation to perform
        :type publish: Publish
        """
        LOGGER.info("publish publish: {}".format(publish))

        self._pending_confirm.update(
            {self._delivery_tag(): publish.attempt()}
        )

        self._channel.basic_publish(
            exchange=publish.topic,
            routing_key="",
            body=publish.message,
        )

        LOGGER.debug("publish pending confirms: {}"
                     .format(self._pending_confirm))

    def rpc_request(self, rpc_request):
        """
        Send an RPC request.

        :param rpc_request: request to send
        :type rpc_request: RPCRequest
        """
        LOGGER.info("rpc_request request: {}".format(rpc_request))

        self._pending_confirm.update(
            {self._delivery_tag(): rpc_request.attempt()}
        )

        self._channel.basic_publish(
            exchange=DEFAULT_EXCHANGE,
            routing_key=rpc_request.receiver,
            properties=BasicProperties(
                correlation_id=rpc_request.correlation_id,
                reply_to=rpc_request.reply_to
            ),
            body=rpc_request.message,
        )

        LOGGER.debug(f"rpc_request pending confirms: {self._pending_confirm}")

    def rpc_response(self, rpc_response):
        """
        Send an RPC response.

        :param rpc_response: RPC response to send
        :type rpc_response: RPCResponse
        """
        LOGGER.info(f"rpc_response response: {rpc_response}")

        self._pending_confirm.update(
            {self._delivery_tag(): rpc_response.attempt()}
        )

        self._channel.basic_publish(
            exchange=DEFAULT_EXCHANGE,
            routing_key=rpc_response.receiver,
            properties=BasicProperties(
                correlation_id=rpc_response.correlation_id
            ),
            body=rpc_response.message,
        )

        LOGGER.debug(f"rpc_response pending confirms: {self._pending_confirm}")

    def command(self, command):
        """
        Send a Command.

        :param command: command work item
        :type command: Command
        """
        LOGGER.info(f"command: {command}")

        self._pending_confirm.update(
            {self._delivery_tag(): command.attempt()}
        )

        self._channel.basic_publish(
            exchange=DEFAULT_EXCHANGE,
            routing_key=command.command_queue,
            body=command.command
        )

        LOGGER.debug(f"command pending confirms: {self._pending_confirm}")

    def _delivery_tag(self):
        """
        Safely steps the delivery tag by using a Lock().

        :return: new delivery tag
        """
        LOGGER.info(f"_delivery_tag current tag: "
                    f"{self._expected_delivery_tag}")

        self._delivery_tag_lock.acquire()
        self._expected_delivery_tag = self._expected_delivery_tag + 1
        delivery_tag = self._expected_delivery_tag

        LOGGER.info(f"_delivery_tag new tag: {self._expected_delivery_tag}")

        self._delivery_tag_lock.release()

        return delivery_tag

    def on_delivery_confirmed(self, frame):
        """
        Callback for when a publish is confirmed

        :param pika.frame.Method frame: message frame, either a Basic.Ack or
                                        Basic.Nack
        """
        LOGGER.info("on_delivery_confirmed frame: {}".format(frame))

        confirm = self._pending_confirm.pop(frame.method.delivery_tag)

        if isinstance(frame.method, Basic.Nack):
            # Retry!
            LOGGER.error("on_delivery_confirmed delivery Nacked!")
            self.handle_work(confirm)

        LOGGER.debug(f"on_delivery_confirmed pending confirms: "
                     f"{self._pending_confirm}")



import signal
import logging

from threading import Thread

from .consumer_channel import RMQConsumerChannel
from .connection import RMQConnection
from rabbitmq_client import log


LOGGER = logging.getLogger(__name__)


def create_consumer_connection(connection_parameters,
                               work_queue,
                               consumed_messages,
                               log_queue,
                               log_level):
    """
    Interface function to instantiate and connect a consumer connection. This
    function is intended as a target for a new process to avoid having to
    instantiate the RMQConsumerConnection outside of the new process' memory
    context.

    :param connection_parameters: pika.ConnectionParameters or None
    :param IPCQueue work_queue: process shared queue used to issue work for the
                                consumer connection
    :param IPCQueue consumed_messages: process shared queue used to forward
                                       messages received for a subscribed topic
                                       to the controlling process
    :param log_queue: queue to post log writes to
    :param log_level: log level
    """
    # Configure logging
    log.set_process_log_handler(log_queue, log_level)

    consumer_connection = RMQConsumerConnection(connection_parameters,
                                                work_queue,
                                                consumed_messages)
    consumer_connection.connect()


class RMQConsumerConnection(RMQConnection):
    """
    Specific connection implementation for RMQ Consumers. This class ensures:
    1. handling of channels according to what is needed for a RabbitMQ
       consumer.
    2. listening on a process shared work queue where work can be posted
       towards the RMQ consumer connection, for example to subscribe to a new
       topic.
    3. posting consumed messages to a process shared queue so that incoming
       messages can be read and handled by the controlling process.
    """

    def __init__(self,
                 connection_parameters,
                 work_queue,
                 consumed_messages):
        """
        Initializes the RMQConsumerConnection with two queues and binds signal
        handlers. The two queues are used to communicate between the connection
        and controlling process. The work queue can be used to issue commands,
        and the consumed messages queue is used to forward received messages to
        the controlling process.

        :param connection_parameters: pika.ConnectionParameters or None
        :param work_queue: process shared queue used to issue work for the
                           consumer connection
        :param consumed_messages: process shared queue used to forward messages
                                  received for a subscribed topic to the
                                  controlling process
        """
        LOGGER.debug("__init__")

        self._work_queue = work_queue

        self._channel = RMQConsumerChannel(consumed_messages)
        self._work_thread = Thread(target=self.monitor_work_queue, daemon=True)

        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.terminate)

        super().__init__(connection_parameters)

    def on_connection_open(self, connection):
        """
        Callback when a connection has been established to the RMQ server.

        :param pika.SelectConnection connection: established connection
        """
        LOGGER.info("on_connection_open connection: {}".format(connection))

        self._channel.open_channel(connection, self.on_channel_open)

    def on_connection_closed(self, _connection, reason):
        """
        Callback for when a connection is closed.

        :param _connection: connection that was closed
        :param reason: reason for closing
        """
        log_string = "on_connection_closed connection: {} reason: {}" \
            .format(_connection, reason)

        if self._closing:
            LOGGER.info(log_string)
            self.finalize_disconnect()
        else:
            # TODO: reconnect handling goes here
            LOGGER.critical(log_string)
            self.finalize_disconnect()

    def on_channel_open(self):
        """
        Callback for notifying a channel has been completely established.
        """
        LOGGER.debug("on_channel_open")

        self.consumer_connection_started()

    def consumer_connection_started(self):
        """
        Shall be called when the consumer connection has been established to
        the point where it is ready to receive work from its controlling
        process. In this case, when a channel has been established.
        """
        LOGGER.info("consumer_connection_started")

        self._work_thread.start()

    def monitor_work_queue(self):
        """
        NOTE!

        This function should live in its own thread so that the
        RMQConsumerConnection is able to respond to incoming work as quickly as
        possible.

        Monitors the consumer connection's work queue and executes from it as
        soon as work is available.
        """
        while True:
            LOGGER.debug("monitor_work_queue waiting for work")

            # Blocking, set block=false to not block
            consume = self._work_queue.get()
            LOGGER.debug("got work to do")
            self._channel.handle_consume(consume)

    def interrupt(self, _signum, _frame):
        """
        Signal handler for signal.SIGINT.

        :param int _signum: signal.SIGINT
        :param ??? _frame: current stack frame
        """
        LOGGER.debug("interrupt")
        self.stop()

    def terminate(self, _signum, _frame):
        """
        Signal handler for signal.SIGTERM.

        :param int _signum: signal.SIGTERM
        :param ??? _frame: current stack frame
        """
        LOGGER.debug("terminate")
        self.stop()

    def stop(self):
        """General stop handler function."""
        self.disconnect()



from rabbitmq_client.common_defs import Printable


"""
Exchange properties
"""


DEFAULT_EXCHANGE = ""


"""
Produce work item types
"""


class GenProduce(Printable):
    """
    Base for producing work classes.
    """
    MAX_ATTEMPTS = 3

    def __init__(self):
        """"""
        self.attempts = 0

    def attempt(self):
        """
        Increments the attempt counter.
        """
        self.attempts += 1

        return self


class Publish(GenProduce):
    """
    A Publish work item.
    """

    def __init__(self, topic, message):
        """
        :param topic: publishing topic
        :param bytes message: message to publish
        """
        self.topic = topic
        self.message = message
        super().__init__()


class RPCRequest(GenProduce):
    """
    An RPC request work item.
    """

    def __init__(self, receiver, message, correlation_id, reply_to):
        """
        :param receiver: receiver of the RPC request
        :param message: content
        :param correlation_id: ID of request
        :param reply_to: reply queue for answer
        """
        self.receiver = receiver
        self.message = message
        self.correlation_id = correlation_id
        self.reply_to = reply_to
        super().__init__()


class RPCResponse(GenProduce):
    """
    An RPC response work item.
    """

    def __init__(self, receiver, message, correlation_id):
        """
        :param receiver: sender of an RPC request
        :param message: content
        :param correlation_id: ID of request that response is for
        """
        self.receiver = receiver
        self.message = message
        self.correlation_id = correlation_id
        super().__init__()


class Command(GenProduce):
    """
    A Command work item.
    """

    def __init__(self, command_queue, command):
        """
        :param command_queue: name of the command queue to send the command to
        :type command_queue: str
        :param command: command to send to command queue
        :type command: bytes
        """
        self.command_queue = command_queue
        self.command = command
        super().__init__()
